

/****************************************************************************
*                   HELP.H - Version 1.0 - 1 July 1994
*
*  This module forms part of the POV help generation system.
*
*  from Persistence of Vision Raytracer
*  Copyright 1994 Persistence of Vision Team
*  Copyright 1994 Christopher J. Cason.
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided so that users may experiment
*  with enhancements to POV-Ray and to port the software to platforms other
*  than those supported by the POV-Ray Team.  There are strict rules under
*  which you are permitted to use this file.  The rules are in the file
*  named POVLEGAL.DOC which should be distributed with this file. If
*  POVLEGAL.DOC is not available or for more info please contact the POV-Ray
*  Team Coordinator by leaving a message in CompuServe's Graphics Developer's
*  Forum.  The latest version of POV-Ray may be found there as well.
*
*  POV-Ray files may also be obtained from ftp.uwa.edu.au in pub/povray.
*
*  This program was written in its entirety by Christopher J. Cason.
*  Its use is freely and permanently granted to the POV-Team and POV users
*  under the conditions specified in POVLEGAL.DOC.
*
*  Author : C. J. Cason (cjcason@yarrow.wt.uwa.edu.au, CIS 100032,1644)
*
*****************************************************************************/

#define VERSION         100

#define RIGHT_SHIFT     0x01
#define SHIFT           0x02
#define CONTROL         0x04
#define ALT             0x08

#define HIGHLIGHT_ON    0xb0
#define HIGHLIGHT_OFF   0xb1
#define BOLD_ON         0xb2
#define BOLD_OFF        0xb3
#define CODE_ON         0xb4
#define CODE_OFF        0xb5
#define HEADING_ON      0xb6
#define HEADING_OFF     0xb7
#define LINE_ON         0xb8
#define LINE_OFF        0xb9
#define PRESERVE_ON     0xba
#define PRESERVE_OFF    0xbb
#define TABLE_ON        0xbc
#define TABLE_OFF       0xbd
#define REFERENCE_ON    0xbe
#define REFERENCE_OFF   0xbf
#define LIST_ON         0xc0
#define LIST_OFF        0xc1
#define LIST_ENTRY_ON   0xc2
#define LIST_ENTRY_OFF  0xc3
#define ITALIC_ON       0xc4
#define ITALIC_OFF      0xc5
#define UNDERSCORE_ON   0xc6
#define UNDERSCORE_OFF  0xc7

#define PARAGRAPH       0xf0
#define INDENT          0xf1
#define REFERENCE       0xf2
#define TARGET          0xf3
#define ESCAPE          0xfe

#define CODE_FRAGMENT   0x01
#define HEADING         0x02
#define LINE_DRAWING    0x04
#define LIST_ENTRY      0x08
#define BOLD            0x10
#define HIGHLIGHT       0x20
#define REF             0x40

typedef struct
{
  char                  signature [8] ;
  unsigned short        reader_version ;
  unsigned short        database_version ;
  char                  sec_number_len ;
  char                  sec_title_len ;
  unsigned long         flags ;
  unsigned short        author_count ;
  unsigned short        section_count ;
  unsigned short        appendix_count ;
  unsigned short        reference_count ;
  unsigned long         title ;
  unsigned long         title_length ;
  unsigned long         copyright ;
  unsigned long         copyright_length ;
  unsigned long         authors ;
  unsigned long         author_length ;
  unsigned long         table_of_contents ;
  unsigned long         table_of_contents_length ;
  unsigned long         reference_index ;
  unsigned long         reference_index_length ;
  unsigned long         reference_string_length ;
  unsigned long         references ;
  unsigned long         reference_length ;
} help_file_header ;

typedef struct _text_struct
{
  struct _text_struct   *next ;
  unsigned long         text ;
} text_struct ;

typedef struct _TOC_struct
{
  struct _TOC_struct    *next ;
  char                  section_number [16] ;
  char                  title [81] ;
  char                  is_appendix ;
  long                  offset ;
  unsigned long         section ;
  unsigned long         section_length ;
} TOC_struct ;

typedef struct
{
  unsigned long         section ;
  unsigned long         section_length ;
  char                  is_appendix ;
  char                  spare ;
} TOC_entry ;

typedef struct _reference_struct
{
  struct _reference_struct        *next ;
  struct _reference_struct        *child ;
  char                            *subject ;
  char                            priority ;
  unsigned                        targets ;
  unsigned                        child_count ;
  unsigned                        refno ;
  unsigned long                   location ;
} reference_struct ;

typedef struct _reference_request
{
  struct _reference_request       *next ;
  char                            *subject ;
  unsigned long                   location ;
} reference_request ;